@@ -13,17 +13,18 @@
 typedef enum { LEFT_EDGE, TOP_EDGE, RIGHT_EDGE, BOTTOM_EDGE } Edge;
 
 typedef struct {
-    double visual_bell_duration, cursor_blink_interval, cursor_stop_blinking_after, mouse_hide_wait, click_interval, wheel_scroll_multiplier;
+    double visual_bell_duration, cursor_blink_interval, cursor_stop_blinking_after, mouse_hide_wait, click_interval, wheel_scroll_multiplier, touch_scroll_multiplier;
     bool enable_audio_bell;
     CursorShape cursor_shape;
     unsigned int open_url_modifiers;
     unsigned int rectangle_select_modifiers;
     unsigned int url_style;
+    unsigned int scrollback_pager_history_size;
     char_type select_by_word_characters[256]; size_t select_by_word_characters_count;
     color_type url_color, background, active_border_color, inactive_border_color, bell_border_color;
     double repaint_delay, input_delay;
-    bool focus_follows_mouse;
-    bool macos_option_as_alt, macos_hide_titlebar, macos_hide_from_tasks, x11_hide_window_decorations, macos_quit_when_last_window_closed, macos_window_resizable, macos_traditional_fullscreen;
+    bool focus_follows_mouse, hide_window_decorations;
+    bool macos_option_as_alt, macos_hide_from_tasks, macos_quit_when_last_window_closed, macos_window_resizable, macos_traditional_fullscreen;
     float macos_thicken_font;
     int adjust_line_height_px, adjust_column_width_px;
     float adjust_line_height_frac, adjust_column_width_frac;
@@ -51,6 +52,7 @@ typedef struct {
 typedef struct {
     double at;
     int button, modifiers;
+    double x, y;
 } Click;
 
 #define CLICK_QUEUE_SZ 3
@@ -66,7 +68,10 @@ typedef struct {
     CursorShape last_cursor_shape;
     PyObject *title;
     ScreenRenderData render_data;
-    unsigned int mouse_cell_x, mouse_cell_y;
+    struct {
+        unsigned int cell_x, cell_y;
+        double x, y;
+    } mouse_pos;
     WindowGeometry geometry;
     ClickQueue click_queue;
     double last_drag_scroll_at;
@@ -97,6 +102,7 @@ typedef struct {
     bool is_set;
 } OSWindowGeometry;
 
+enum WAYLAND_RENDER_STATE { RENDER_FRAME_NOT_REQUESTED, RENDER_FRAME_REQUESTED, RENDER_FRAME_READY };
 
 typedef struct {
     void *handle;
@@ -126,7 +132,7 @@ typedef struct {
     FONTS_DATA_HANDLE fonts_data;
     id_type temp_font_group_id;
     double pending_scroll_pixels;
-    unsigned int nsgl_ctx_updated;
+    enum WAYLAND_RENDER_STATE wayland_render_state;
 } OSWindow;
 
 
@@ -138,11 +144,12 @@ typedef struct {
     OSWindow *os_windows;
     size_t num_os_windows, capacity;
     OSWindow *callback_os_window;
-    bool close_all_windows;
+    bool terminate;
     bool is_wayland;
     bool debug_gl, debug_font_fallback;
     bool has_pending_resizes;
     bool in_sequence_mode;
+    bool tab_bar_hidden;
     double font_sz_in_pts;
     struct { double x, y; } default_dpi;
     id_type active_drag_in_window;
@@ -172,6 +179,7 @@ void event_loop_wait(double timeout);
 void swap_window_buffers(OSWindow *w);
 void make_window_context_current(OSWindow *w);
 void hide_mouse(OSWindow *w);
+bool is_mouse_hidden(OSWindow *w);
 void destroy_os_window(OSWindow *w);
 void focus_os_window(OSWindow *w, bool also_raise);
 void set_os_window_title(OSWindow *w, const char *title);
@@ -201,4 +209,6 @@ typedef enum {
 } CocoaPendingAction;
 void set_cocoa_pending_action(CocoaPendingAction action);
 bool application_quit_requested();
+void request_application_quit();
 #endif
+void wayland_request_frame_render(OSWindow *w);
